package com.unitesk.requality.api;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.annotation.Annotation;
import java.lang.reflect.Type;
import java.util.Arrays;

import javax.ws.rs.Consumes;
import javax.ws.rs.DELETE;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.PUT;
import javax.ws.rs.Path;
import javax.ws.rs.Produces;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.container.ContainerRequestContext;
import javax.ws.rs.container.ContainerRequestFilter;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.MultivaluedMap;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.Status;
import javax.ws.rs.ext.ExceptionMapper;
import javax.ws.rs.ext.MessageBodyReader;
import javax.ws.rs.ext.MessageBodyWriter;
import javax.ws.rs.ext.Provider;

import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.module.SimpleModule;
import com.fasterxml.jackson.databind.module.SimpleSerializers;
import com.fasterxml.jackson.databind.ser.std.StdSerializer;
import com.unitesk.requality.api.models.JsonError;
import com.unitesk.requality.api.models.JsonException;
import com.unitesk.requality.api.models.JsonInputBase;
import com.unitesk.requality.api.models.JsonOutputBase;
import com.unitesk.requality.api.models.JsonResponseWrapper;
import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.RLogger;

@Produces(MediaType.APPLICATION_JSON)
public abstract class BaseService {
  static class ResponseWrapperSerializer extends StdSerializer<JsonResponseWrapper> {
    private static final long serialVersionUID = 1L;

    public ResponseWrapperSerializer() {
      this(JsonResponseWrapper.class);
    }

    public ResponseWrapperSerializer(Class<JsonResponseWrapper> clazz) {
      super(clazz);
    }

    @Override
    public void serialize(JsonResponseWrapper obj, JsonGenerator ret, SerializerProvider prov)
        throws IOException {

      ret.writeStartObject();
      ret.writeFieldName("success");

      if (obj.hasError()) {
        ret.writeBoolean(false);
        ret.writeFieldName("error");
      } else {
        ret.writeBoolean(true);
        ret.writeFieldName("result");
      }

      if (obj.hasObject()) {
        ret.writeObject(obj.getObject());
      } else {
        ret.writeTree(obj.getRawJson());
      }

      ret.writeEndObject();
    }
  }

  @Provider
  @Produces(MediaType.APPLICATION_JSON)
  @Consumes(MediaType.APPLICATION_JSON)
  static class JsonWriter implements MessageBodyWriter<Object>, MessageBodyReader<Object> {
    private ObjectMapper mapper = new ObjectMapper();

    public JsonWriter() {
      mapper.enable(SerializationFeature.INDENT_OUTPUT);
      mapper.disable(SerializationFeature.FAIL_ON_EMPTY_BEANS);
      mapper.registerModule(
          new SimpleModule() {
            private static final long serialVersionUID = 1L;

            @Override
            public void setupModule(SetupContext ctx) {
              super.setupModule(ctx);

              ctx.addSerializers(
                  new SimpleSerializers(Arrays.asList(new ResponseWrapperSerializer())));
            }
          });
    }

    @Override
    public long getSize(Object arg0, Class<?> arg1, Type arg2, Annotation[] arg3, MediaType arg4) {
      return -1;
    }

    @Override
    public boolean isWriteable(Class<?> arg0, Type arg1, Annotation[] arg2, MediaType arg3) {
      return arg3.equals(MediaType.APPLICATION_JSON_TYPE)
          && (JsonOutputBase.class.isAssignableFrom(arg0) || JsonNode.class.isAssignableFrom(arg0));
    }

    @Override
    public void writeTo(
        Object arg0,
        Class<?> arg1,
        Type arg2,
        Annotation[] arg3,
        MediaType arg4,
        MultivaluedMap<String, Object> arg5,
        OutputStream arg6)
        throws IOException, WebApplicationException {
      mapper.writeValue(arg6, new JsonResponseWrapper(arg0));
    }

    @Override
    public boolean isReadable(Class<?> arg0, Type arg1, Annotation[] arg2, MediaType arg3) {
      return arg3.equals(MediaType.APPLICATION_JSON_TYPE)
          && (JsonInputBase.class.isAssignableFrom(arg0) || JsonNode.class.isAssignableFrom(arg0));
    }

    @Override
    public Object readFrom(
        Class<Object> arg0,
        Type arg1,
        Annotation[] arg2,
        MediaType arg3,
        MultivaluedMap<String, String> arg4,
        InputStream arg5)
        throws IOException, WebApplicationException {
      return mapper.readValue(arg5, arg0);
    }
  }


  public static class ExceptionHandler implements ExceptionMapper<Exception> {
    @Override
    public Response toResponse(Exception arg0) {
      RLogger.logInfo(CoreUtils.getTrace(arg0));
      if(arg0 instanceof
          JsonException)
      {
        JsonError err = ((JsonException) arg0).getError();
        return Response.status(err.getHttpStatus()).entity(err).build();
      }
      JsonError err = new JsonError(arg0.getMessage(), CoreUtils.getTrace(arg0));
      return Response.status(err.getHttpStatus()).entity(err).build();
    }
  }

  public static class SecurityFilter implements ContainerRequestFilter {
    private JsonError error =
        new JsonError(
            Status.UNAUTHORIZED, "unauthorized", "You are not authorized to perform this request");

    private String apiKey;

    public SecurityFilter(String apiKey) {
      this.apiKey = apiKey;
    }

    public void setApiKey(String apiKey) {
      this.apiKey = apiKey;
    }

    @Override
    public void filter(ContainerRequestContext arg0) throws IOException {
      String authorization = arg0.getHeaderString("Authorization");
      if (apiKey == null || authorization == null || !authorization.equals("Bearer " + apiKey)) {
        throw error.exception();
      }
    }
  }

  @Path("{dummy:.+}")
  public static class FallbackService extends BaseService {
    private JsonError error =
        new JsonError(Status.NOT_FOUND, "not_found", "No such method was found");

    @GET
    public Object return404Get() throws JsonException {
      throw error.exception();
    }

    @POST
    public Object return404Post() throws JsonException {
      throw error.exception();
    }

    @PUT
    public Object return404Put() throws JsonException {
      throw error.exception();
    }

    @DELETE
    public Object return404Delete() throws JsonException {
      throw error.exception();
    }
  }
}
