package com.unitesk.requality.api.services.report;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.util.UUID;
import java.util.stream.Collectors;

import javax.ws.rs.DELETE;
import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.core.Response.Status;

import com.fasterxml.jackson.databind.node.TextNode;
import com.unitesk.requality.api.BaseService;
import com.unitesk.requality.api.Utils;
import com.unitesk.requality.api.models.IDList;
import com.unitesk.requality.api.models.IDMapping;
import com.unitesk.requality.api.models.JsonEmpty;
import com.unitesk.requality.api.models.JsonError;
import com.unitesk.requality.api.models.JsonException;
import static com.unitesk.requality.api.services.project.ProjectService.getProject;

import com.unitesk.requality.core.CoreUtils;
import com.unitesk.requality.core.ResManager;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.nodetypes.ReportData;
import com.unitesk.requality.report.Template;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

@Path("/projects/{p}/reports")
public class ReportService extends BaseService {
  public static final JsonError REPORT_NOT_FOUND =
      new JsonError(Status.NOT_FOUND, "report_not_found", "Specified report does not exists");

  @GET
  public IDList httpList(@PathParam("p") String project) throws JsonException {
    TreeDB db = getProject(project);
    TreeNode docRoot = db.getNode("/Reports");
    return new IDList(
        Utils.stream(new DeepFirstTreeWalker(docRoot, false, true, true).iterator())
            .filter(x -> x.getType().equals(ReportData.TYPE_NAME))
            .map(IDMapping::fromNode)
            .collect(Collectors.toList()));
  }

  @GET
  @Path("/{r}")
  public TextNode httpContents(@PathParam("p") String project, @PathParam("r") UUID repId)
      throws JsonException {
    ReportData report = getReport(project, repId);
    Template tpl = Activator.getDefault().getReportTemplates().get(report.getTemplateId());

    URI savePath = report.getSaveDir();
    String href = null;
    if (tpl == null) {
      if (report.getTemplateId() != null) {
        String lastGen = report.attrs().getStringAttributeValue(ReportData.ATTR_LAST_GENERATED);
        if (lastGen != null)
          try {
            href = Activator.getUrl(new URI(savePath + CoreUtils.URLEncode(lastGen))).toString();
          } catch (IOException e) {
            e.printStackTrace();
            throw JsonError.INTERNAL_ERROR.exception();
          } catch (URISyntaxException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
          }
      } else {
        tpl = Activator.getDefault().getReportTemplates().get(Template.DEFAULT_ID);
      }
    }

    if (tpl != null) {
      href =
          savePath
              + File.separator
              + tpl.getTemplates().get(0).getOutputFileName(report.getTreeDB());
    }

    if (href == null) {
      throw JsonError.INTERNAL_ERROR.exception();
    }

    try {
      File f = new File(href);

      if (!f.exists() || !f.isFile() || !f.canRead()) {
        throw JsonError.INTERNAL_ERROR.exception();
      }

      String r = new String(Files.readAllBytes(f.toPath()), StandardCharsets.UTF_8);
      return new TextNode(r);
    } catch (IOException e) {
      e.printStackTrace();
      throw JsonError.INTERNAL_ERROR.exception();
    }
  }

  @DELETE
  @Path("/{r}")
  public JsonEmpty httpDelete(@PathParam("p") String project, @PathParam("r") UUID repId)
      throws JsonException {
    getReport(project, repId).delete();
    return JsonEmpty.INSTANCE;
  }

  public static ReportData getReport(TreeDB db, UUID uuid) throws JsonException {
    TreeNode node = db.getNode(uuid);

    if (node == null || !node.getType().equals(ReportData.TYPE_NAME))
      throw REPORT_NOT_FOUND.exception();

    return (ReportData) node;
  }

  public static ReportData getReport(String project, UUID repId) throws JsonException {
    return getReport(getProject(project), repId);
  }
}
